#include "imu_measure.h"

namespace Localization {
namespace Ego {

double ImuMeasure::latestUpdateTime_ = -1.0;

ImuMeasure::ImuMeasure(const MeasurementType &mType, const ImuInfo &imuInfo)
:BaseMeasure(mType)
{
  isImuBlock_ = false;
  stamp_ = imuInfo.stamp_;
  if (stamp_ - latestUpdateTime_ < 1e-5) {
    isImuBlock_ = true;
    std::cout<<std::setprecision(15)<<"ImuMeasure():this frame is repeated,stamp"<<stamp_
        <<",latestUpdateTime_"<<latestUpdateTime_<<std::endl;
  }
  latestUpdateTime_ = stamp_;

  seq_ = imuInfo.seq_;
  type_ = mType;
  isKFUsed_ = false;

  acc_ = imuInfo.acc_;
  gyro_ = imuInfo.gyro_;
  mRm_ = Eigen::MatrixXd::Zero(6, 6);
  mRm_(0, 0) = imuAccNoiseVar_;
  mRm_(1, 1) = imuAccNoiseVar_;
  mRm_(2, 2) = 2.0 * 2.0 * imuAccNoiseVar_;
  mRm_.block<3, 3>(3, 3) = imuGyroNoiseVar_ * Eigen::Matrix3d::Identity();
  CheckValid();
};

void ImuMeasure::CheckValid()
{
  isImuValid_ = true;
  if (fabs(lastImuStamp_ - lastVehStamp_) > 0.5) {
    std::cerr<<std::setprecision(15)<<"ImuMeasure():time error,imu stamp has big diff with veh stamp,lastImuStamp:"<<
         lastImuStamp_<<",lastVehStamp_:"<<lastVehStamp_<<",curImuStamp:"<<stamp_<<std::endl;
    isImuBlock_ = true;
  } else {
    isImuBlock_ = false;
  }
  lastImuStamp_ = stamp_;

  if (isRtkValid_ == false) {
    mRm_.block<3, 3>(3, 3) = 0.5 * imuGyroNoiseVar_ * Eigen::Matrix3d::Identity();
    std::cout<<std::setprecision(15)<<"imu-CheckValid():isRtkValid_ false,use smaller imu noise,stamp:"<<stamp_<<std::endl;
  }
}

bool ImuMeasure::Update(const State &state, State &nextState)
{
  nextState = state;
  if (isImuValid_ == false) {
    std::cout<<"imu-Update():isImuValid_ false,can't update,use next pridict"<<std::endl;
    return false;
  }
  nextState.time = stamp_;

  Eigen::VectorXd innov = Eigen::VectorXd::Zero(6);
  Eigen::MatrixXd mH = Eigen::MatrixXd::Zero(6, StateIndex::STATE_TOTAL);
  ObserveIMU(state,nextState,innov,mH);
  (void)UpdateStateIMU(state, nextState, innov, mH);
  return true;
}

bool ImuMeasure::UpdateStateIMU(const State &state, State &nextState, const Eigen::VectorXd &innov,
     const Eigen::MatrixXd &mH)
{
  Eigen::MatrixXd mS = mH * state.mP * mH.transpose() + mRm_;
  //==todo:S矩阵求逆降算力
  Eigen::MatrixXd mK = state.mP * mH.transpose() * mS.inverse();
  Eigen::VectorXd dx = mK * innov;
  dx.segment<3>(StateIndex::R) = Eigen::Vector3d::Zero();
  dx.segment<3>(StateIndex::P) = Eigen::Vector3d::Zero();
  dx.segment<3>(StateIndex::V) = Eigen::Vector3d::Zero();
  dx.segment<3>(StateIndex::BA) = Eigen::Vector3d::Zero();
  dx.segment<3>(StateIndex::BG) = Eigen::Vector3d::Zero();
  dx.segment<3>(StateIndex::IMU_INSTALL_ANGLE) = Eigen::Vector3d::Zero();
  dx(StateIndex::WS) = 0.0;

  UpdateState(state, nextState, dx, mH, mK);

  return true;
}

Eigen::VectorXd ImuMeasure::MeasurementEquation(const State &state) const
{
  Eigen::VectorXd measurement(6);
  measurement.head(3) = state.q.inverse() * (state.acc - gravity_) + state.ba;
  measurement.tail(3) = state.av + state.bg;
  return measurement;
}

void ImuMeasure::ObserveIMU(const State &state, State &nextState, Eigen::VectorXd &innov,
                                            Eigen::MatrixXd &mH)
{
  innov.head(3) =
      acc_ - (state.q.inverse() * (state.acc - gravity_) + state.ba);
  innov.tail(3) = gyro_ - (state.av + state.bg);

  Eigen::Matrix3d mI = Eigen::Matrix3d::Identity();

  mH.block<3, 3>(0, StateIndex::R) =
      SkewSymmetric(state.q.inverse() * (state.acc - gravity_));
  mH.block<3, 3>(0, StateIndex::ACC) =
      state.q.toRotationMatrix().transpose();
  mH.block<3, 3>(3, StateIndex::AV) = mI;
  Eigen::Matrix3d tmp;
  mH.block<3, 3>(0, StateIndex::BA) = tmp.setZero();
  mH.block<3, 3>(3, StateIndex::BG) = tmp.setZero();
}

} // namespace Ego
} // namespace Localizaiton

